import timeit
from typing import Callable

import numpy as np
from matplotlib.figure import Figure

from diffeq_solver_tk.differential_equation_metadata import WaveEquationMetadata, BoundaryType
from diffeq_solver_tk.services.differential_equation_service import BoundedEquationService


class WaveEquationService(BoundedEquationService):
    metadata: WaveEquationMetadata

    def __init__(self, main_figure: Figure):
        super().__init__(main_figure)

    def compute_solution(self, metadata: WaveEquationMetadata) -> np.ndarray:
        start_time = timeit.default_timer()
        initial_values: Callable[[float], float] = lambda x: eval(metadata.initial_values)
        initial_derivatives: Callable[[float], float] = lambda x: eval(metadata.initial_derivatives)
        source: Callable[[float, float], float] = lambda t, x: eval(metadata.source)
        N = metadata.samples
        dx = metadata.length / (N - 1)
        K = 2 * int(metadata.wave_speed * metadata.time / dx) + 1
        dt = metadata.time / (K - 1)
        D = self.create_time_step_matrix((metadata.wave_speed * dt / dx) ** 2, N)
        left_values: Callable[[float], float] = lambda t: eval(metadata.boundary_conditions.left.values)
        right_values: Callable[[float], float] = lambda t: eval(metadata.boundary_conditions.right.values)
        is_left_dirichlet = metadata.boundary_conditions.left.type == BoundaryType.DIRICHLET
        is_right_dirichlet = metadata.boundary_conditions.right.type == BoundaryType.DIRICHLET

        # Initialize solution array
        solution = np.zeros((K, N))
        solution[0] = initial_values(np.arange(N) * dx)
        solution[1, 1:N - 1] = (1 / 2) * D @ solution[0]
        solution[1, 1:N - 1] += dt ** 2 * initial_derivatives(np.arange(1, N - 1) * dx)
        solution[1, 1:N - 1] += (dt ** 2 / 2) * source(dt, np.arange(1, N - 1) * dx)
        solution[1, 0] = left_values(dt) if is_left_dirichlet else solution[1, 1] - left_values(dt) * dx
        solution[1, -1] = right_values(dt) if is_right_dirichlet else solution[1, -2] + right_values(dt) * dx
        for k in range(2, K):
            T = k * dt
            solution[k, 1:N - 1] = D @ solution[k - 1] - solution[k - 2, 1:N - 1]
            solution[k, 1:N - 1] += dt ** 2 * source(T, np.arange(1, N - 1) * dx)
            solution[k, 0] = left_values(T) if is_left_dirichlet else solution[k, 1] - left_values(T) * dx
            solution[k, -1] = right_values(T) if is_right_dirichlet else solution[k, -2] + right_values(T) * dx

        total_time = timeit.default_timer() - start_time
        print(f"computing wave equation solution took {total_time} seconds")
        return solution

    def validate_metadata(self, metadata: WaveEquationMetadata):
        if metadata.samples <= 1 or metadata.length == 0 or metadata.time == 0:
            raise Exception("invalid wave equation metadata")

    def create_time_step_matrix(self, r: float, N: int) -> np.ndarray:
        D = np.zeros((N - 2, N))
        for i in range(N - 2):
            D[i, i] = r
            D[i, i + 1] = 2 * (1 - r)
            D[i, i + 2] = r
        return D



from tkinter import Frame, Entry, Variable, Label, StringVar, Radiobutton, Button
from typing import Generic, TypeVar, Union, Callable, Optional, Dict

import diffeq_solver_tk.messages.common_messages as common_messages
import diffeq_solver_tk.tkinter_config as config
from diffeq_solver_tk.differential_equation_metadata import BoundaryType

T = TypeVar('T')


class EquationFormBuilder(Generic[T]):
    __field_entry_map: Dict[T, Union[Entry, Variable]]

    def __init__(self, frame: Frame):
        self.frame = frame
        self.__field_entry_map = dict()

    def build_entry_row(self, field: T, display_name: str, symbol_name: str, row: int):
        self.__field_entry_map[field] = Entry(master=self.frame, font=config.details_font, width=24)
        self.__place_label(f"{display_name}:", row, 0, 18, "w")
        self.__place_label(f"{symbol_name} = ", row, 1, 0, "e")
        self.__field_entry_map[field].grid(row=row, column=2, columnspan=2)

    def build_boundary_type_section(self, left_boundary_type_field: T, right_boundary_type_field: T):
        self.__place_label(common_messages.left_boundary_type, 0, 0, 18, "w")
        self.__build_boundary_type_row(left_boundary_type_field, 0)
        self.__place_label(common_messages.right_boundary_type, 1, 0, 18, "w")
        self.__build_boundary_type_row(right_boundary_type_field, 1)

    def create_button(self, text: str, callback: Optional[Callable[[], None]] = None) -> Button:
        button = Button(
            master=self.frame,
            text=text,
            font=config.details_font,
            width=10,
        )
        if callback is not None:
            button.configure(command=callback)
        return button

    def get_field_entry_map(self) -> Dict[T, Union[Entry, Variable]]:
        return self.__field_entry_map

    def __place_label(self, text: str, row: int, column: int, horizontal_padding: int, sticky: str):
        Label(master=self.frame,
              text=text,
              font=config.details_font,
              background=config.details_background).grid(
            row=row, column=column, padx=horizontal_padding, pady=6, sticky=sticky)

    def __build_boundary_type_row(self, boundary_type_field: T, row: int):
        self.__field_entry_map[boundary_type_field] = StringVar(value=BoundaryType.DIRICHLET.value)
        self.__place_radio_button(common_messages.dirichlet,
                                  self.__field_entry_map[boundary_type_field],
                                  BoundaryType.DIRICHLET.value,
                                  row, 1)
        self.__place_radio_button(common_messages.neumann,
                                  self.__field_entry_map[boundary_type_field],
                                  BoundaryType.NEUMANN.value,
                                  row, 2)

    def __place_radio_button(self, text: str, variable: Variable, value: str, row: int, column: int):
        Radiobutton(master=self.frame,
                    text=text,
                    font=config.details_font,
                    background=config.details_background,
                    activebackground=config.details_background,
                    variable=variable,
                    value=value).grid(row=row, column=column, sticky="w")



from abc import abstractmethod, ABC
from typing import TypeVar, Generic, Optional

import numpy as np
from matplotlib import cm
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from openpyxl import Workbook

from diffeq_solver_tk.differential_equation_metadata import OrdinaryDifferentialEquationMetadata, BoundedEquationMetadata

T = TypeVar('T')


class DifferentialEquationService(Generic[T], ABC):
    metadata: Optional[T] = None
    solution: Optional[np.ndarray] = None
    main_figure: Figure

    def __init__(self, main_figure: Figure):
        self.main_figure = main_figure

    def compute_and_update_solution(self, metadata: T):
        self.validate_metadata(metadata)
        self.metadata = metadata
        self.solution = self.compute_solution(metadata)
        self.render_current_solution()

    def clear_solution(self):
        self.metadata = None
        self.solution = None
        self.clear_figure()

    def clear_figure(self):
        self.main_figure.clf()

    @abstractmethod
    def compute_solution(self, metadata: T) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def export_solution(self, table_path: str):
        raise NotImplementedError

    @abstractmethod
    def render_current_solution(self):
        raise NotImplementedError

    @abstractmethod
    def validate_metadata(self, metadata: T):
        raise NotImplementedError


class OrdinaryDifferentialEquationService(DifferentialEquationService[OrdinaryDifferentialEquationMetadata], ABC):
    def __init__(self, main_figure: Figure):
        super().__init__(main_figure)

    def render_current_solution(self):
        domain = np.linspace(0, self.metadata.time, self.metadata.samples)
        ax = self.main_figure.add_subplot()
        ax.plot(domain, self.solution)
        ax.set_xlabel("t [time]")
        ax.set_ylabel("x(t)")
        ax.plot()
        ax.set_title("Solution: x(t)")

    def export_solution(self, table_path: str):
        domain = np.linspace(0, self.metadata.time, self.metadata.samples)
        workbook = Workbook()
        worksheet = workbook.active
        worksheet.cell(1, 1, 't')
        worksheet.cell(1, 2, 'x')
        row = 2
        for t, x in zip(domain, self.solution):
            worksheet.cell(row, 1, t)
            worksheet.cell(row, 2, x)
            row += 1
        workbook.save(table_path)


class BoundedEquationService(DifferentialEquationService[BoundedEquationMetadata], ABC):
    __animation: Optional[FuncAnimation] = None
    __is_animation_playing: bool = False

    def __init__(self, main_figure: Figure):
        super().__init__(main_figure)

    def render_current_solution(self):
        self.clear_figure()
        self.clear_animation()
        K, N = self.solution.shape
        time_domain = np.linspace(0, self.metadata.time, K)
        space_domain = np.linspace(0, self.metadata.length, N)
        x, t = np.meshgrid(space_domain, time_domain)
        ax = self.main_figure.add_subplot(projection='3d')
        ax.set_xlabel('x [length]')
        ax.set_ylabel('t [time]')
        ax.set_zlabel('u(x, t)')
        surf = ax.plot_surface(x, t, self.solution, cmap=cm.coolwarm, linewidth=0, antialiased=False)
        self.main_figure.colorbar(surf, shrink=0.5, aspect=5)

    def generate_animation(self):
        self.clear_figure()
        K, N = self.solution.shape
        ax = self.main_figure.add_subplot()
        Axes.set_xlim(ax, left=0, right=self.metadata.length)
        Axes.set_ylim(ax, bottom=np.min(self.solution), top=np.max(self.solution))
        ax.set_xlabel("x [length]")
        ax.set_ylabel("u(x, t)")
        ax.set_title("Solution: u(x, t)")
        space_domain = np.linspace(0, self.metadata.length, self.metadata.samples)
        line, = ax.plot(space_domain, self.solution[0], "-o", markersize=4)
        time_text = ax.text(0.82, 0.92, '', transform=ax.transAxes)
        dt = self.metadata.time / K

        def update_plot(k: int):
            line.set_ydata(self.solution[k])
            time_text.set_text("t = %.3f" % (k * dt))
            return line, time_text

        self.__animation = FuncAnimation(self.main_figure, update_plot, frames=K, blit=True, interval=20)

    def toggle_animation(self):
        if not self.__animation:
            self.generate_animation()
            self.__is_animation_playing = True
        elif self.is_animation_playing():
            self.__animation.pause()
            self.__is_animation_playing = False
        else:
            self.__animation.resume()
            self.__is_animation_playing = True

    def is_animation_playing(self):
        return self.__is_animation_playing

    def export_solution(self, table_path: str):
        K, N = self.solution.shape
        time_domain = np.linspace(0, self.metadata.time, K)
        space_domain = np.linspace(0, self.metadata.length, N)
        workbook = Workbook()
        worksheet = workbook.active
        worksheet.cell(1, 2, "x →")
        for i in range(N):
            worksheet.cell(1, 3 + i, space_domain[i])
        worksheet.write(2, 1, "t ↓")
        for k in range(K):
            worksheet.write(3 + k, 1, time_domain[k])
        for k in range(K):
            for i in range(N):
                worksheet.write(k + 3, i + 3, self.solution[k, i])
        workbook.save(table_path)

    def clear_animation(self):
        if self.__animation:
            self.__animation.pause()
            self.__animation = None
            self.__is_animation_playing = False
            self.main_figure.clf()



left_boundary_values = "Left boundary values"
right_boundary_values = "Right boundary values"
initial_values = "Initial values"
initial_derivatives = "Initial Derivatives"
source_term = "Source term (optional)"
wave_speed = "Wave speed"
length = "Length"
time_interval = "Time interval"
samples = "Samples"

left_boundary_values_symbol = "φ₁(t)"
right_boundary_values_symbol = "φ₂(t)"
initial_values_symbol = "u(0, x)"
initial_derivatives_symbol = "∂u(0, x)/∂t"
source_term_symbol = "s(t, x)"
wave_speed_symbol = "c"
length_symbol = "L"
time_interval_symbol = "T"
samples_symbol = "N"
